/*
 * Copyright (C) 2007 Rich Waters <raw@goodhumans.com>
 *
 * This file is part of lib_mysqludf_preg.
 *
 * lib_mysqudf_preg is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * lib_mysqludf_preg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


/** @file ghmysql.c
 *  
 * @brief Provides some functions that are shared by the various
 *        gh udf functions.   
 *
 */
#include "ghmysql.h"


/**
 * @fn char *ghstrndup( char *s , int l )
 *
 * @brief safe strndup
 *
 * @param s - bytes to copy
 * @param l - number of bytes to copy
 *
 * @return ptr -  to newly allocate memory 
 * @return NULL - if memory not available
 *
 * @detail - This function copies the given number of bytes from the given
 * buffer into a newly allocated memory area.  It then add a '\0' to the
 * end.  This function shouldn't be necessary, but since many call rely on
 * null-termination instead of lengths, this creates things that can be
 * passed to those functions safely.
 */
char *ghstrndup( char *s , int l )
{
    char *colval = NULL ;

    colval = malloc( l + 1 ) ;
    if( !colval )
    {
        fprintf( stderr , "Not enough memory: %d\n" , l ) ;
        return NULL ;
    }
    memcpy( colval , s , l ) ;
    colval[ l ] = '\0'; 

    return colval ;
}


/**
 * @fn char *ghargdup( UDF_ARGS *args,int i ) 
 *
 * @brief extract argument as a newly allocated null-terminated memory area
 *
 * @param args - args from the UDF function call
 * @param i - offset of argument to extract.
 *
 * @return pointer -  to newly allocate memory if successful
 * @return NULL -  if arg is NULL or not enough memory
 * 
 */
char *ghargdup( UDF_ARGS *args,int i ) 
{      
    int l ;
    char *s ;
    char *colval = NULL ;

    if( (l=args->lengths[i] ) && (s=args->args[i]) )
    {
        colval = ghstrndup( s , l ) ;
    }

    return colval ;
}        

/**
 * @fn char *ghargdups( UDF_ARGS *args,int i ) 
 *
 * @brief extract argument as a newly allocated null-terminated memory area,
 * and set the length to the size of that area
 *
 * @param args - args from the UDF function call
 * @param i - offset of argument to extract.
 *
 * @return pointer -  to newly allocate memory if successful
 *
 * @details This is useful for guaranteeing not to have NULL as an argument
 */
char *ghargdups( UDF_ARGS *args , int i , unsigned long *l)
{
    if( !(*l=args->lengths[ i ]))
    {   
        return strdup( "" ) ;
    }
    else if( !args->args[i]) {
        *l = 0 ; 
        return strdup( "" ) ;
    }
    else {
        return ghargdup( args , i ) ;
    }
}
